export class ParseError extends Error {
    constructor(input, message, index) {
        super(index !== undefined ? `$message at index $index` : message);
        this.input = input;
        this.message = message;
        this.index = index;
    }
}
function modulo(a, n) {
    return (n + (a % n)) % n;
}
function _arrayToOctetString(input) {
    let out;
    out = (input[0] >> 8) + '.' + (input[0] & 0xff) + '.';
    out += (input[1] >> 8) + '.' + (input[1] & 0xff);
    return out;
}
function _isAddr(addr) {
    return addr instanceof Addr;
}
function _arrayToHex(input, zeroElide, zeroPad) {
    let i;
    let elStart = null;
    let elLen = 0;
    if (zeroElide) {
        let start = null;
        let len = null;
        for (i = 0; i < input.length; i++) {
            if (input[i] === 0) {
                if (start === null) {
                    start = i;
                    len = 1;
                }
                else {
                    len++;
                }
            }
            else if (start !== null) {
                if (len > elLen) {
                    elStart = start;
                    elLen = len;
                }
                start = null;
            }
        }
        if (start !== null && len > elLen) {
            elStart = start;
            elLen = len;
        }
    }
    const output = [];
    let num;
    for (i = 0; i < input.length; i++) {
        if (elStart !== null) {
            if (i === elStart) {
                if (elLen === 8) {
                    return ['::'];
                }
                else if (elStart === 0 || elStart + elLen === input.length) {
                    output.push(':');
                }
                else {
                    output.push('');
                }
            }
            if (i >= elStart && i < elStart + elLen) {
                continue;
            }
        }
        num = input[i].toString(16);
        if (zeroPad && num.length !== 4) {
            num = '0000'.slice(num.length) + num;
        }
        output.push(num);
    }
    return output;
}
function _ipv4Mapped(input) {
    const comp = [0, 0, 0, 0, 0, 0xffff];
    let i;
    for (i = 0; i < 6; i++) {
        if (input[i] !== comp[i]) {
            return false;
        }
    }
    return true;
}
class AddrAttrs {
    constructor() {
        this.ipv4Bare = false;
        this.ipv4Mapped = false;
    }
}
export class Addr {
    constructor() {
        this._fields = [0, 0, 0, 0, 0, 0, 0, 0];
        this._attrs = new AddrAttrs();
    }
    getKind() {
        if (v4subnet.contains(this)) {
            return 'ipv4';
        }
        else {
            return 'ipv6';
        }
    }
    toString(opts) {
        const format = opts.format === 'auto' ? this._deriveFormatFromAttrs() : opts.format;
        switch (format) {
            case 'v4':
                if (!v4subnet.contains(this)) {
                    throw new Error('cannot print non-v4 address in dotted quad notation');
                }
                return _arrayToOctetString(this._fields.slice(6));
            case 'v4-mapped':
                if (!v4subnet.contains(this)) {
                    throw new Error('cannot print non-v4 address as a v4-mapped address');
                }
                let output = _arrayToHex(this._fields.slice(0, 6), opts.zeroElide === undefined ? true : opts.zeroElide, opts.zeroPad || false);
                output.push(_arrayToOctetString(this._fields.slice(6)));
                return output.join(':');
            case 'v6':
                return _arrayToHex(this._fields, opts.zeroElide === undefined ? true : opts.zeroElide, opts.zeroPad || false).join(':');
            default:
                throw new Error('unrecognized format method "' + format + '"');
        }
    }
    compare(b) {
        let i;
        for (i = 0; i < 8; i++) {
            if (this._fields[i] < b._fields[i]) {
                return -1;
            }
            else if (this._fields[i] > b._fields[i]) {
                return 1;
            }
        }
        return 0;
    }
    and(val) {
        const input = _isAddr(val) ? val : Addr.toAddr(val);
        const output = this.clone();
        for (let i = 0; i < 8; i++) {
            output._fields[i] = output._fields[i] & input._fields[i];
        }
        return output;
    }
    clone() {
        const out = new Addr();
        out._fields = this._fields.slice();
        for (let k in this._attrs) {
            out._attrs[k] = this._attrs[k];
        }
        return out;
    }
    not() {
        var output = this.clone();
        for (let i = 0; i < 8; i++) {
            output._fields[i] = (~output._fields[i]) & 0xffff;
        }
        return output;
    }
    or(input) {
        const addr = Addr.toAddr(input);
        const output = this.clone();
        for (let i = 0; i < 8; i++) {
            output._fields[i] = output._fields[i] | addr._fields[i];
        }
        return output;
    }
    offset(num) {
        if (num < -4294967295 || num > 4294967295) {
            throw new Error('offsets should be between -4294967295 and 4294967295');
        }
        const out = this.clone();
        let i, moved;
        for (i = 7; i >= 0; i--) {
            moved = out._fields[i] + num;
            if (moved > 65535) {
                num = moved >>> 16;
                moved = moved & 0xffff;
            }
            else if (moved < 0) {
                num = Math.floor(moved / (1 << 16));
                moved = modulo(moved, 1 << 16);
            }
            else {
                num = 0;
            }
            out._fields[i] = moved;
            if (num !== 0) {
                if ((i === 0) || (i === 6 && this._attrs.ipv4Mapped)) {
                    return null;
                }
            }
            else {
                break;
            }
        }
        return out;
    }
    static toAddr(input) {
        if (typeof (input) === 'string') {
            return Addr._ip6AddrParse(input);
        }
        else if (_isAddr(input)) {
            return input;
        }
        else {
            throw new Error('Invalid argument: Addr or parsable string expected');
        }
    }
    static _ip6AddrParse(input) {
        input = input.toLowerCase();
        let result = new Addr();
        let ip6Fields = [];
        let ip4Fields = [];
        let expIndex = null;
        let value = '';
        let i, c;
        if (input.length > 40) {
            throw new ParseError(input, 'Input too long');
        }
        for (i = 0; i < input.length; i++) {
            c = input[i];
            if (c === ':') {
                if ((i + 1) < input.length && input[i + 1] === ':') {
                    if (expIndex !== null) {
                        throw new ParseError(input, 'Multiple :: delimiters', i);
                    }
                    if (value !== '') {
                        ip6Fields.push(value);
                        value = '';
                    }
                    expIndex = ip6Fields.length;
                    i++;
                }
                else {
                    if (value === '') {
                        throw new ParseError(input, 'illegal delimiter', i);
                    }
                    ip6Fields.push(value);
                    value = '';
                }
            }
            else if (c === '.') {
                ip4Fields.push(value);
                value = '';
            }
            else {
                value = value + c;
            }
        }
        if (value !== '') {
            if (ip4Fields.length !== 0) {
                ip4Fields.push(value);
            }
            else {
                ip6Fields.push(value);
            }
        }
        else {
            if (expIndex !== ip6Fields.length || ip4Fields.length > 0) {
                throw new ParseError(input, 'Cannot end with delimiter besides ::');
            }
        }
        if (ip4Fields.length === 0) {
            if (ip6Fields.length > 8) {
                throw new ParseError(input, 'Too many fields');
            }
            else if (ip6Fields.length < 8 && expIndex === null) {
                throw new ParseError(input, 'Too few fields');
            }
        }
        else {
            if (ip4Fields.length !== 4) {
                throw new ParseError(input, 'IPv4 portion must have 4 fields');
            }
            if (ip6Fields.length === 0 && expIndex === null) {
                result._attrs.ipv4Bare = true;
                ip6Fields = ['ffff'];
                expIndex = 0;
            }
            if (ip6Fields.length > 6) {
                throw new ParseError(input, 'Too many fields');
            }
            else if (ip6Fields.length < 6 && expIndex === null) {
                throw new ParseError(input, 'Too few fields');
            }
        }
        let field, num;
        for (i = 0; i < ip6Fields.length; i++) {
            field = ip6Fields[i];
            num = parseInt(field, 16);
            if (num instanceof Error || num < 0 || num > 65535) {
                throw new ParseError(input, 'Invalid field value: ' + field);
            }
            ip6Fields[i] = num;
        }
        for (i = 0; i < ip4Fields.length; i++) {
            field = ip4Fields[i];
            num = parseInt(field, 10);
            if (num instanceof Error || num < 0 || num > 255) {
                throw new ParseError(input, 'Invalid field value: ' + field);
            }
            ip4Fields[i] = num;
        }
        if (ip4Fields.length !== 0) {
            ip6Fields.push((ip4Fields[0] * 256) + ip4Fields[1]);
            ip6Fields.push((ip4Fields[2] * 256) + ip4Fields[3]);
        }
        if (ip6Fields.length < 8 && expIndex !== null) {
            let filler = [];
            for (i = 0; i < (8 - ip6Fields.length); i++) {
                filler.push(0);
            }
            ip6Fields = Array.prototype.concat(ip6Fields.slice(0, expIndex), filler, ip6Fields.slice(expIndex));
        }
        if (ip4Fields.length !== 0) {
            if (!_ipv4Mapped(ip6Fields)) {
                throw new ParseError(input, 'invalid dotted-quad notation');
            }
            else {
                result._attrs.ipv4Mapped = true;
            }
        }
        result._fields = ip6Fields;
        return result;
    }
    _deriveFormatFromAttrs() {
        if (this._attrs.ipv4Bare) {
            return 'v4';
        }
        else if (this._attrs.ipv4Mapped) {
            return 'v4-mapped';
        }
        return 'v6';
    }
}
export class CIDR {
    constructor(addr, prefixLen) {
        if (prefixLen === undefined) {
            let fields = addr.match(/^([a-fA-F0-9:.]+)\/([0-9]+)$/);
            if (fields === null) {
                throw new Error('Invalid argument: <addr>/<prefix> expected');
            }
            addr = fields[1];
            prefixLen = parseInt(fields[2], 10);
        }
        prefixLen = prefixLen || 0;
        this._addr = Addr.toAddr(addr);
        if (this._addr._attrs.ipv4Bare) {
            prefixLen += 96;
        }
        if (prefixLen < 0 || prefixLen > 128) {
            throw new Error('Invalid prefix length');
        }
        this._prefix = prefixLen;
        this._mask = this._prefixToAddr(prefixLen);
        this._addr = this._addr.and(this._mask);
    }
    contains(input) {
        const addr = Addr.toAddr(input);
        return (this._addr.compare(addr.and(this._mask)) === 0);
    }
    first() {
        if (this._prefix >= 127) {
            return this._addr;
        }
        else {
            return this._addr.offset(1);
        }
    }
    last() {
        let ending = this._addr.or(this._mask.not());
        if (this._prefix >= 127) {
            return ending;
        }
        else {
            if (this._addr._attrs.ipv4Mapped) {
                return ending.offset(-1);
            }
            else {
                return ending;
            }
        }
    }
    compare(b) {
        const cmp = this._addr.compare(b._addr);
        return cmp === 0 ? b._prefix - this._prefix : cmp;
    }
    prefixLength(format) {
        if (format === undefined || format === 'auto') {
            format = this._addr._attrs.ipv4Bare ? 'v4' : 'v6';
        }
        switch (format) {
            case 'v4':
                if (!v4subnet.contains(this._addr)) {
                    throw new Error('cannot return v4 prefix length for non-v4 address');
                }
                return this._prefix - 96;
            case 'v6':
                return this._prefix;
            default:
                throw new Error('unrecognized format method "' + format + '"');
        }
    }
    get address() {
        return this._addr;
    }
    toString(opts) {
        var _a;
        let format = (_a = opts === null || opts === void 0 ? void 0 : opts.format) !== null && _a !== void 0 ? _a : 'auto';
        if (format === 'v4-mapped') {
            format = 'v6';
        }
        return this._addr.toString(opts) + '/' + this.prefixLength(format);
    }
    _prefixToAddr(len) {
        if (len > 128 || len < 0) {
            throw new Error('$len is out of bounds 0-128');
        }
        let output = new Addr();
        let i;
        for (i = 0; len > 16; i++, len -= 16) {
            output._fields[i] = 0xffff;
        }
        if (len > 0) {
            output._fields[i] = 0xffff - ((1 << (16 - len)) - 1);
        }
        return output;
    }
}
const v4subnet = new CIDR('::ffff:0:0', 96);
//# sourceMappingURL=ip6addr.js.map